"""Read & extract service metadata and its available operations"""
import re

import boto3


def parse_resource_and_operations(
    hub, service_name: str, session: "boto3.session.Session"
):
    """
    Get resource and their available operations for client initialized for a given service

    @returns
        Mapping of resource to its methods and corresponding boto3 operation_name.
        {
            "resource": {
                { "method" : "operation_name" }
            }
        }
    """
    operations = {}
    client = session.client(service_name=service_name, region_name="us-west-2")

    for op in client.meta.method_to_api_mapping:
        try:
            verb, resource = op.split("_", maxsplit=1)
            if re.match(rf"\w+[^aoius]s$", resource):
                resource = hub.tool.format.inflect.singular(resource)
            # Special case for resource names that end with apis
            if resource.endswith("apis"):
                resource = resource[:-1]
            if resource not in operations:
                operations[resource] = {}

            if op.endswith(f"{resource}s") and verb.startswith("get"):
                operations[resource]["list"] = op

            if not operations.get(resource, {}).get(verb):
                # do not replace it
                # TODO: Is there a better way to get right methods for a resource?
                operations[resource][verb] = op
        except ValueError:
            hub.log.error("Failure in extracting operation metadata")

    return operations


def parse_docstring(hub, session: "boto3.session.Session", service_name: str):
    """
    Get service description
    """
    client = session.client(service_name=service_name, region_name="us-west-2")
    plugin_docstring = hub.tool.format.html.parse(client._service_model.documentation)
    return "\n".join(hub.tool.format.wrap.wrap(plugin_docstring, width=120))


def parse_service_tag_methods(
    hub, session: "boto3.session.Session", aws_service_name: str
):
    """
    Parses service tag method definitions. There is usually a common method at service level which can be used for
    tagging. Sometimes it is single method for update and sometimes there are separate add/remove methods.
    Capture them all here.
    """
    tag_methods = dict()
    try:
        client = session.client(service_name=aws_service_name, region_name="us-west-2")
        for op in client.meta.method_to_api_mapping:
            if re.match("(add|create|put|tag).*_(resource|tags|tagging)", op):
                tag_methods["tag_resource"] = hub.pop_create.aws.function.parse(
                    session=session,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            elif re.match("(remove|delete|untag).*_(resource|tags|tagging)", op):
                tag_methods["untag_resource"] = hub.pop_create.aws.function.parse(
                    session=session,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            elif re.match("(list|get|describe).*_(tags|tagging)", op):
                tag_methods["list_tags"] = hub.pop_create.aws.function.parse(
                    session=session,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            elif re.match("(change).*_(tags)", op):
                tag_methods["update_tags"] = hub.pop_create.aws.function.parse(
                    session=session,
                    aws_service_name=aws_service_name,
                    resource_name=None,
                    func_name=op,
                )
            else:
                continue
    except Exception as err:
        hub.log.error(
            f"Error when generating tag action definitions for {aws_service_name}: {err.__class__.__name__}: {err}"
        )

    return tag_methods



"""Create tool related plugins e.g. conversion_utils"""
from typing import Any
from typing import Dict


def parse(
    hub,
    shared_resource_data: dict,
) -> Dict[str, Any]:
    aws_service_name = shared_resource_data.get("aws_service_name")
    resource_name = shared_resource_data.get("resource_name")

    plugin = {
        "doc": f"Utils related functions for {aws_service_name}.{resource_name}.\n",
        "imports": [
            "import copy",
            "from typing import *",
        ],
        "functions": {
            "convert_raw_resource_to_present_async": hub.pop_create.aws.utils.plugin.generate_convert_raw_to_present(
                resource_name, shared_resource_data["get"]
            ),
        },
    }

    return plugin


def generate_convert_raw_to_present(
    hub, resource_name: str, get_function_definition: dict
):
    return {
        "doc": f"Convert raw resource of {resource_name} type into present format.\n",
        "params": dict(
            idem_resource_name=hub.pop_create.aws.utils.template.NAME_PARAMETER.copy(),
            resource_id=hub.pop_create.aws.utils.template.RESOURCE_ID_PARAMETER.copy(),
            raw_resource=hub.pop_create.aws.utils.template.RAW_RESOURCE_PARAMETER.copy(),
        ),
        "hardcoded": dict(
            **get_function_definition.get("hardcoded", {}),
        ),
    }



"""Parse tags plugin"""
from dict_tools.data import NamespaceDict


def parse(hub, aws_service_name: str, tag_functions: dict):
    """
    Parse tag plugin for the service
    """
    plugin = {
        "doc": f"Tag related functions for AWS service '{aws_service_name}'.",
        "imports": [
            "import copy",
            "from typing import *",
        ],
        "functions": NamespaceDict(
            get_tags_for_resource=hub.pop_create.aws.tag.plugin.generate_get_tags_for_resource(
                tag_functions
            ),
            update_tags=hub.pop_create.aws.tag.plugin.generate_update_tags(
                tag_functions
            ),
        ),
    }

    return plugin


def generate_get_tags_for_resource(hub, tag_functions):
    list_tags_function_definition = tag_functions.get("list_tags", {})
    return {
        "doc": f"Get tags for a given resource.\n",
        "params": dict(
            resource_id=hub.pop_create.aws.tag.template.RESOURCE_ID_PARAMETER.copy(),
        ),
        "hardcoded": dict(
            **list_tags_function_definition.get("hardcoded", {}),
        ),
    }


def generate_update_tags(hub, tag_functions):
    update_tags_function_definition = tag_functions.get("update_tags", {})
    if update_tags_function_definition:
        return {
            "doc": f"Updates tags for a given resource.\n",
            "params": update_tags_function_definition["params"],
            "hardcoded": dict(
                update_tags_boto3_function=update_tags_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_function"),
                update_tags_input_params=update_tags_function_definition.get(
                    "params", {}
                ).keys(),
                single_update=True,
            ),
        }
    else:
        untag_resource_function_definition = tag_functions.get("untag_resource", {})
        tag_resource_function_definition = tag_functions.get("tag_resource", {})
        return {
            "doc": f"Updates tags for a given resource.\n",
            "params": dict(
                resource_id=hub.pop_create.aws.tag.template.RESOURCE_ID_PARAMETER.copy(),
                old_tags=hub.pop_create.aws.tag.template.OLD_TAGS_PARAMETER.copy(),
                new_tags=hub.pop_create.aws.tag.template.NEW_TAGS_PARAMETER.copy(),
            ),
            "hardcoded": dict(
                untag_resource_boto3_function=untag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_function"),
                untag_resource_input_params=untag_resource_function_definition.get(
                    "params", {}
                ).keys(),
                untag_resource_documentation=untag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_documentation"),
                tag_resource_boto3_function=tag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_function"),
                tag_resource_input_params=tag_resource_function_definition.get(
                    "params", {}
                ).keys(),
                tag_resource_documentation=tag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_documentation"),
                single_update=False,
            ),
        }



"""Exec functions for AWS API Gateway v2 Domain Name resources."""
from typing import Any
from typing import Dict


async def get(hub, ctx, name, resource_id: str) -> Dict[str, Any]:
    """Gets a domain name.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            The domain name.

    Returns:
        Dict[str, Any]:
            Returns a domain name.

    Examples:
        Call from the CLI:

        .. code-block:: bash

            idem exec aws.apigatewayv2.domain_name.get name="my_resource" resource_id="domain_name"

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, name:str, resource_id:str):
                await hub.exec.aws.apigatewayv2.domain_name.get(ctx, name, resource_id)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigatewayv2.domain_name.get
                - kwargs:
                    name: my_resource
                    resource_id: domain_name
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.apigatewayv2.get_domain_name(
        ctx, DomainName=resource_id
    )
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.domain_name", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = hub.tool.aws.apigatewayv2.domain_name.convert_raw_domain_name_to_present(
        raw_resource=ret["ret"]
    )
    return result



"""Build plugin metadata which can be used by pop-create for plugin code generation"""
from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict


def parse_exec_plugin(
    hub,
    ctx,
    shared_resource_data: dict,
) -> Dict[str, Any]:
    aws_service_name = shared_resource_data.get("aws_service_name")
    resource_name = shared_resource_data.get("resource_name")

    plugin = {
        "doc": f"Exec module for managing {aws_service_name}.{resource_name} in AWS.\nService methods can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/{aws_service_name}.html",
        "imports": [
            "import copy",
            "from dataclasses import field",
            "from dataclasses import make_dataclass",
            "from typing import *",
        ],
        "func_alias": {"list_": "list"},
        "functions": NamespaceDict(
            get=hub.pop_create.aws.plugin.generate_get(shared_resource_data),
            list=hub.pop_create.aws.plugin.generate_list(shared_resource_data),
            create=hub.pop_create.aws.plugin.generate_present(shared_resource_data),
            update=hub.pop_create.aws.plugin.generate_update(shared_resource_data),
            delete=hub.pop_create.aws.plugin.generate_absent(shared_resource_data),
        ),
    }

    if ctx.create_plugin == "auto_state":
        plugin["contracts"] = ["auto_state", "soft_fail"]

    return plugin


def parse_state_plugin(
    hub,
    ctx,
    shared_resource_data: dict,
) -> Dict[str, Any]:
    aws_service_name = shared_resource_data.get("aws_service_name")
    resource_name = shared_resource_data.get("resource_name")

    plugin = {
        "doc": f"State module for managing {aws_service_name}.{resource_name} in AWS.\n Service methods can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/{aws_service_name}.html",
        "imports": [
            "import copy",
            "from dataclasses import field",
            "from dataclasses import make_dataclass",
            "from typing import *",
        ],
        "contracts": ["resource"],
        "functions": NamespaceDict(
            present=hub.pop_create.aws.plugin.generate_present(shared_resource_data),
            absent=hub.pop_create.aws.plugin.generate_absent(shared_resource_data),
            describe=hub.pop_create.aws.plugin.generate_list(shared_resource_data),
        ),
    }

    return plugin


def generate_list(hub, shared_resource_data):
    describe_function_definition = shared_resource_data["list"]
    params = describe_function_definition.get("params", {})
    return {
        "doc": f"{describe_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **describe_function_definition.get("hardcoded", {}),
        ),
    }


def generate_present(hub, shared_resource_data):
    create_function_definition = shared_resource_data["create"]
    params = create_function_definition.get("params", {})
    if "Tags" not in params.keys():
        params["tags"] = hub.pop_create.aws.template.TAGS_PARAMETER.copy()
    return {
        "doc": f"{create_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **create_function_definition.get("hardcoded", {}),
        ),
    }


def generate_absent(hub, shared_resource_data):
    delete_function_definition = shared_resource_data["delete"]
    # In most cases, absent/delete only needs name and resource_id params
    # This will make sure that all these params are not added in function definition
    return {
        "doc": f"{delete_function_definition.get('doc', '')}",
        "params": delete_function_definition.get("params", {}),
        "hardcoded": dict(
            **delete_function_definition.get("hardcoded", {}),
        ),
    }


def generate_get(hub, shared_resource_data):
    get_function_definition = shared_resource_data["get"]
    params = get_function_definition.get("params", {})
    return {
        "doc": f"\n{get_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **get_function_definition.get("hardcoded", {}),
        ),
    }


def generate_update(hub, shared_resource_data):
    update_function_definition = shared_resource_data["update"]
    params = update_function_definition.get("params", {})
    if "tags" not in params.keys():
        params["tags"] = hub.pop_create.aws.template.TAGS_PARAMETER.copy()
    return {
        "doc": f"{update_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **update_function_definition.get("hardcoded", {}),
        ),
    }



"""List of possible functions for known operations."""

DESCRIBE_FUNCTIONS = ("get", "search", "describe")

LIST_FUNCTIONS = ("list", "describe", "search")

DELETE_FUNCTIONS = (
    "delete",
    "disassociate",
    "reject",
    "deallocate",
    "unassign",
    "deregister",
    "deprovision",
    "revoke",
    "release",
    "terminate",
    "cancel",
    "disable",
)

CREATE_FUNCTIONS = (
    "create",
    "associate",
    "accept",
    "allocate",
    "assign",
    "register",
    "provision",
    "authorize",
    "run",
    "enable",
    "upload",
    "put",
    "publish",
    "request",
    "put",
    "add",
)

UPDATE_FUNCTIONS = (
    "modify",
    "update",
)



NAME_PARAMETER = {
    "default": None,
    "doc": "An Idem name of the resource",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

RESOURCE_ID_PARAMETER = {
    "default": None,
    "doc": "An identifier of the resource in the provider",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

RAW_RESOURCE_PARAMETER = {
    "default": None,
    "doc": "The raw representation of the resource in the provider",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

CONVERT_RAW_RESOURCE_TO_PRESENT_REQUEST = """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id
    }

    resource_parameters = OrderedDict(
    {{ function.hardcoded.raw_resource_to_resource_mapping|pprint|indent(12,true) }}
    )

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    # Get it from raw_resource or explicitly retrieved tags
    resource_tags_list = raw_resource.get("Tags") or raw_resource.get("TagList")

    if resource_tags_list:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            resource_tags_list
        )
    else:
        resource_translated["tags"] = await hub.tool.aws.{{ function.hardcoded.aws_service_name }}.tag.get_tags_for_resource(
            ctx,
            resource_id=resource_id
        )

    return resource_translated
"""



TAGS_PARAMETER = {
    "default": None,
    "doc": "The tags to apply to the resource. Defaults to None.",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

PRESENT_REQUEST_FORMAT = """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs") and v is not None
    }

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before_ret = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )
        if not before_ret["result"] or not before_ret["ret"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

        result["old_state"] = copy.deepcopy(before_ret["ret"])
    if before:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state=desired_state
            )
            result["comment"] = (f"Would update aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} '{name}'",)
            return result

        # TODO: Add other required parameters (including tags, if necessary)
        update_ret = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.update(
            ctx,
            name=name,
            resource_id=resource_id,
            # TODO: Add other required parameters (including tags, if necessary): **{{ parameter.mapping.kwargs|default({}) }}
        )

        resource_updated = bool(update_ret["ret"])
        if not resource_updated:
            result["comment"] += (f"'{name}' already exists",)
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        result["comment"] += (
            f"Updated aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} '{name}'",
        )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state=desired_state
            )
            result["comment"] = (f"Would create aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} {name}",)
            return result

        create_ret = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.create(
            ctx,
            name=name,
            resource_id=resource_id,
            # TODO: Add other required parameters from: **{{ parameter.mapping.kwargs|default({})}}
        )
        result["result"] = create_ret["result"]
        if not result["result"]:
            result["comment"] += (create_ret["comment"],)
            return result

        result["comment"] += (f"Created '{name}'",)

        # TODO: extract resource_id from create_ret
        resource_id = create_ret["ret"]["TODO: extract resource_id from the response"]
        # This makes sure the created resource is saved to esm regardless if the subsequent update call fails or not.
        result["new_state"] = {"name": name, "resource_id": resource_id}
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}", name=name
        )

    # TODO: Add other required parameters
    # Possible parameters: **{{ parameter.mapping.kwargs|default({}) }}
    after = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )
    result["new_state"] = after["ret"]
    return result
"""

ABSENT_REQUEST_FORMAT = """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # This is to make absent idempotent. If absent is run again, it would be a no-op
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}", name=name
        )
        return result

    # TODO: Add other required parameters
    # Possible parameters: **{{ parameter.mapping.kwargs|default({}) }}
    before_ret = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    # Case: Error
    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    # Case: Not Found
    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}", name=name
        )
        return result

    if ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = (f"Would delete aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} '{name}'",)
        return result

    result["old_state"] = before_ret["ret"]

    # TODO: Add other required parameters
    # Possible parameters: **{{ parameter.mapping.kwargs|default({}) }}
    delete_ret = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.delete(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    result["result"] = delete_ret["result"]
    if not result["result"]:
        result["comment"] += (delete_ret["comment"],)
        result["result"] = False
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}", name=name
    )
    return result
"""

DESCRIBE_REQUEST_FORMAT = """
    result = {}

    # TODO: Add other required parameters from: {{ parameter.mapping.kwargs|default({}) }}
    ret = await hub.exec.aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.list(
        ctx
    )

    if not ret or not ret["result"]:
        hub.log.debug(f"Could not describe aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} {ret['comment']}")
        return result

    for resource in ret["ret"]:
        # TODO: Look for respective identifier in **{{ function.hardcoded.resource_attributes }}
        resource_id = resource.get("TODO: Replace with resource identifier")
        result[resource_id] = {
            "aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result
"""

GET_REQUEST_FORMAT = """
    result = dict(comment=[], ret=None, result=True)

    # TODO: Map `resource_id` to correct identifier. Check boto3 documentation: {{ function.hardcoded.boto3_documentation }}
    get = await {{ function.hardcoded.boto3_function }}(
        ctx=ctx,
        **{{ parameter.mapping.kwargs|default({}) }}
    )

    # Case: Error
    if not get["result"]:
        # Do not return success=false when it is not found.
        # Most of the resources would return "*NotFound*" type of exception when it is 404
        # TODO: Check NotFound type of exception in boto3 documentation: {{ function.hardcoded.boto3_documentation }}
        if "NotFound" in str(get["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}",
                    name=resource_id
                )
            )
            result["comment"].append(get["comment"])
            return result

        result["comment"].append(get["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    # DOC: Check boto3 documentation for response syntax: {{ function.hardcoded.boto3_documentation }}
    if not get["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}",
                name=resource_id
            )
        )
        return result

    # Case: One matching record is found (If more than one is found, then taking first)
    raw_resource = get["ret"]

    # DOC: Check boto3 documentation for response syntax: {{ function.hardcoded.boto3_documentation }}
    # Resource attributes: {{ function.hardcoded.resource_attributes }}
    result["ret"] = await hub.tool.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}_utils.convert_raw_{{ function.hardcoded.resource_name }}_to_present_async(
        resource_id=resource_id,
        raw_resource=raw_resource,
        tags={},
        idem_resource_name=name,
    )

    return result
"""

LIST_REQUEST_FORMAT = """
    result = dict(comment=[], ret=[], result=True)
    # TODO: Change function methods params if needed
    # DOC: Check boto3 documentation for request syntax: {{ function.hardcoded.boto3_documentation }}
    ret = await {{ function.hardcoded.boto3_function }}(
        ctx=ctx,
        **{{ parameter.mapping.kwargs|default({}) }}
    )

    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result

    # DOC: Check boto3 documentation for response syntax: {{ function.hardcoded.boto3_documentation }}
    if not ret["ret"].get("{{ function.hardcoded.response_key }}"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}", name=None
            )
        )
        return result

    # boto3 documentation: {{ function.hardcoded.boto3_documentation }}
    for resource in ret["ret"]["{{ function.hardcoded.response_key }}"]:
        # Resource attributes: {{ function.hardcoded.resource_attributes }}
        result["ret"].append(await hub.tool.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}_utils.convert_raw_{{ function.hardcoded.resource_name }}_to_present_async(
            resource_id="TODO: map resource id to correct identifier",
            raw_resource=resource,
            tags={},
            idem_resource_name="TODO: map name to correct response param",
        ))
    return result
"""

CREATE_REQUEST_FORMAT = """
    result = dict(comment=[], ret=[], result=True)

    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )

    # TODO: Change function methods params if needed.
    # DOC: Check boto3 documentation for request syntax: {{ function.hardcoded.boto3_documentation }}
    ret = await {{ function.hardcoded.boto3_function }}(
        ctx,
        {{ "ClientToken=name," if function.hardcoded.has_client_token -}}
        **{{ parameter.mapping.kwargs|default({}) }}
    )

    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result

    result["comment"].append(f"Created aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} '{name}'",)

    # DOC: Check response syntax in boto3 documentation: {{ function.hardcoded.boto3_documentation }}
    resource_id = result["ret"]["TODO: <resource-id-key>"]

    # TODO: If tags are added as part of create, the following is redundant
    tag_resource_ret = await hub.tool.aws.{{ function.hardcoded.aws_service_name }}.tag.update_tags(
        ctx,
        resource_id=resource_id,
        old_tags={},
        new_tags=tags
    )
    if not tag_resource_ret["result"]:
        result["result"] = False
        result["comment"].append(tag_resource_ret["comment"])

    return result
"""

UPDATE_REQUEST_FORMAT = """
    result = dict(comment=[], ret=[], result=True)

    desired_state = {"name": name, "resource_id": resource_id, **kwargs}

    # DOC: Check boto3 documentation for request syntax: {{ function.hardcoded.boto3_documentation }}
    resource_to_raw_input_mapping = OrderedDict(
    {{ function.hardcoded.function_resource_to_input_param_mappings|pprint|indent(12,true) }}
    )

    parameters_to_update = {}
    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            parameters_to_update[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    if parameters_to_update:
        ret = await {{ function.hardcoded.boto3_function }}(
            ctx,
            {{ "ClientToken=name," if function.hardcoded.has_client_token -}}
            **parameters_to_update
        )

        if not ret["result"]:
            result["result"] = False
            result["comment"] += (
                f"Could not update aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} '{name}'",
            )
            result["comment"].append(ret["comment"])
            return result

        result["comment"].append(f"Updated aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }} '{name}'",)

        # TODO: If tags are updated as part of update function then the following tag update is redundant
        get_tags_ret = await hub.tool.aws.{{ function.hardcoded.aws_service_name }}.tag.get_tags_for_resource(
            ctx,
            resource_id=resource_id
        )

        if get_tags_ret["result"]:
            current_tags = get_tags_ret.get("ret", {})
            tags = kwargs.get("tags", {})
            tags = (
                hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
                if isinstance(tags, List)
                else tags
            )
            update_tags_ret = await hub.tool.aws.{{ function.hardcoded.aws_service_name }}.tag.update_tags(
                ctx,
                resource_id=resource_id,
                old_tags=current_tags,
                new_tags=tags
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] += update_tags_ret["comment"]
                return result
    return result
"""

DELETE_REQUEST_FORMAT = """
    result = dict(comment=[], ret=[], result=True)

    # Check boto3 documentation for inputs: {{ function.hardcoded.boto3_documentation }}
    delete_ret = await {{ function.hardcoded.boto3_function }}(
        ctx,
        {{ "ClientToken=name," if function.hardcoded.has_client_token -}}
        **{{ parameter.mapping.kwargs|default({}) }}
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        result["comment"] = delete_ret["comment"]
        result["result"] = False
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.resource_name }}", name=name
    )

    return result
"""



"""Build CRUD function definitions for a resource to be used in resource plugins"""
from typing import Any
from typing import Dict


def parse_actions(
    hub,
    session: "boto3.session.Session",
    aws_service_name: str,
    resource_name: str,
    functions: dict,
) -> Dict[str, Any]:
    resource_methods = dict()
    try:
        resource_methods[
            "get"
        ] = hub.pop_create.aws.resource.parse_func_definition_with_possible_names(
            session=session,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_type="describe",
            possible_function_names=hub.pop_create.aws.possible_functions.DESCRIBE_FUNCTIONS,
            aws_functions=functions,
        )

        resource_methods[
            "list"
        ] = hub.pop_create.aws.resource.parse_func_definition_with_possible_names(
            session=session,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_type="describe",
            possible_function_names=hub.pop_create.aws.possible_functions.LIST_FUNCTIONS,
            aws_functions=functions,
        )

        resource_methods[
            "create"
        ] = hub.pop_create.aws.resource.parse_func_definition_with_possible_names(
            session=session,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_type="create",
            possible_function_names=hub.pop_create.aws.possible_functions.CREATE_FUNCTIONS,
            aws_functions=functions,
        )

        resource_methods[
            "delete"
        ] = hub.pop_create.aws.resource.parse_func_definition_with_possible_names(
            session=session,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_type="delete",
            possible_function_names=hub.pop_create.aws.possible_functions.DELETE_FUNCTIONS,
            aws_functions=functions,
        )

        resource_methods[
            "update"
        ] = hub.pop_create.aws.resource.parse_func_definition_with_possible_names(
            session=session,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_type="update",
            possible_function_names=hub.pop_create.aws.possible_functions.UPDATE_FUNCTIONS,
            aws_functions=functions,
        )
    except Exception as err:
        hub.log.error(
            f"Error when generating resource's action definitions for {resource_name}: {err.__class__.__name__}: {err}"
        )

    return resource_methods


def parse_func_definition_with_possible_names(
    hub,
    session: "boto3.session.Session",
    aws_service_name: str,
    resource_name: str,
    func_type: str,
    possible_function_names: list,
    aws_functions: dict,
) -> Dict[str, Any]:
    """
    Create function definitions with possible function names
    """
    possible_func = None
    for func_name in possible_function_names:
        if func_name in aws_functions:
            possible_func = func_name

    if possible_func is None:
        hub.log.info(
            f"Cannot determine function for {func_type} in {aws_service_name}.{resource_name}: {list(aws_functions.keys())}"
        )
        return {}

    return hub.pop_create.aws.function.parse(
        session, aws_service_name, resource_name, aws_functions[possible_func]
    )


def build_resource_init_call(hub, aws_service_name, resource_name):
    """This is generic resource initialization call"""
    return f"resource = await hub.tool.boto3.resource.create(ctx, {aws_service_name}, {resource_name}, resource_id)"


def build_get_by_resource_call(hub):
    """This is to get a resource when generic resource initialization is used"""
    return "await hub.tool.boto3.resource.describe(resource)"



NAME_PARAMETER = {
    "default": None,
    "doc": "An Idem name of the resource",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

RESOURCE_ID_PARAMETER = {
    "default": None,
    "doc": "An identifier of the resource in the provider",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

OLD_TAGS_PARAMETER = {
    "default": None,
    "doc": "Dict in the format of {tag-key: tag-value}",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

NEW_TAGS_PARAMETER = {
    "default": None,
    "doc": "Dict in the format of {tag-key: tag-value}",
    "param_type": "dict",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

GET_TAGS_FOR_RESOURCE_REQUEST = """
    result = dict(comment=[], result=True, ret=None)

    if not resource_id:
        result["result"] = False
        result["comment"] = ["resource_id parameter is None"]
        return result

    # DOC: Check boto3 documentation for request syntax: {{ function.hardcoded.boto3_documentation }}
    tags_ret = await hub.exec.boto3.client.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.boto3_function }}(
        ctx,
        # TODO: Map input params with: {{ function.hardcoded.function_input_param_names }}
    )

    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] = tags_ret["comment"]
        return result

    # TODO: Check boto3 documentation for response syntax: {{ function.hardcoded.boto3_documentation }}
    tags = tags_ret.get("ret").get("TODO: <response_key>") if tags_ret.get("result") else None
    result["ret"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    return result
"""

UPDATE_TAGS_REQUEST = """
    result = dict(comment=[], result=True, ret=None)

    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )

    if (not tags_to_remove) and (not tags_to_add):
        # If there is nothing to add or remove, return from here with old tags, if present
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    {% if function.hardcoded.single_update %}
        # DOC: Convert tag dict to list with `hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add)`
        # TODO: Check boto3 documentation for request syntax: {{ function.hardcoded.update_tags_documentation }}
        change_tags_ret = await hub.exec.boto3.client.{{ function.hardcoded.aws_service_name }}.{{ function.hardcoded.update_tags_boto3_function }}(
            ctx,
            # TODO: Map input params with: {{ function.hardcoded.update_tags_input_params }}
        )

        if not change_tags_ret["result"]:
            result["comment"] = change_tags_ret["comment"]
            result["result"] = False
            return result
    {% else %}
    if tags_to_remove:
        if not ctx.get("test", False):
            # DOC: Check boto3 documentation for response syntax: {{ function.hardcoded.untag_resource_documentation }}
            # TODO: Map {{ function.hardcoded.untag_resource_boto3_function }} input params from: {{ function.hardcoded.untag_resource_input_params }}
            delete_ret = await {{ function.hardcoded.untag_resource_boto3_function }}(
                ctx,
                # TODO: Map input params with: {{ function.hardcoded.untag_resource_input_params }}
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            # DOC: Check boto3 documentation for response syntax: {{ function.hardcoded.tag_resource_documentation }}
            # TODO: Convert tag dict to list with `hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add)`
            add_ret = await {{ function.hardcoded.tag_resource_boto3_function }}(
                ctx,
                # TODO: Map input params with: {{ function.hardcoded.tag_resource_input_params }}
            )
            if not add_ret["result"]:
                result["comment"] += add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    {% endif %}
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result
"""



import pathlib

import boto3.session
from dict_tools.data import NamespaceDict

try:
    import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def __init__(hub):
    hub.pop.sub.load_subdirs(hub.pop_create.aws)


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx, directory)
    ctx.servers = [None]

    # AWS already has an acct plugin
    ctx.has_acct_plugin = False
    ctx.service_name = "aws_auto"

    # Now start getting into AWS resource plugin creation
    session = boto3.session.Session()

    # If CLI provides services then use those services first
    # e.g. --services rds
    services = hub.OPT.pop_create.services or session.get_available_services()
    # This takes a while because we are making http calls to aws
    for aws_service_name in tqdm.tqdm(services, desc="services"):
        # Clean out the service name
        aws_service_name = (
            aws_service_name.lower().strip().replace(" ", "_").replace("-", "_")
        )
        aws_service_name = hub.tool.format.keyword.unclash(aws_service_name)

        # Get supported operations for this service
        resource_operations = hub.pop_create.aws.service.parse_resource_and_operations(
            service_name=aws_service_name,
            session=session,
        )

        requested_service_resources = hub.OPT.pop_create.service_resources
        if bool(requested_service_resources):
            # if the CLI provides resources, filter the list of resources to process
            # e.g. --service_resources db_cluster db_instance
            resource_operations = {
                r: resource_operations[r]
                for r in requested_service_resources
                if r in resource_operations
            }

        for resource_name, functions in tqdm.tqdm(
            resource_operations.items(), desc="operations"
        ):
            resource_modules = dict()
            # Clean out resource name
            resource_name = (
                resource_name.lower().strip().replace(" ", "_").replace("-", "_")
            )

            # Check if the plugin should be created
            #   - see if it exists
            #   - or --overwrite flag is used
            resource_plugin_exists = hub.pop_create.aws.init.plugin_exists(
                ctx, aws_service_name, resource_name
            )
            should_create_resource_plugin = (
                ctx.overwrite_existing or not resource_plugin_exists
            )

            if should_create_resource_plugin:
                # parse known or commonly used resource actions for the resource
                resource_actions = hub.pop_create.aws.resource.parse_actions(
                    session,
                    aws_service_name,
                    resource_name,
                    functions,
                )

                # create shared resource data to be used when creating resource plugins
                shared_resource_data = {
                    "aws_service_name": aws_service_name,
                    "resource_name": resource_name,
                    # CRUD functions
                    "get": resource_actions.get("get"),
                    "create": resource_actions.get("create"),
                    "update": resource_actions.get("update"),
                    "delete": resource_actions.get("delete"),
                    "list": resource_actions.get("list"),
                    "generic_resource_init_call": hub.pop_create.aws.resource.build_resource_init_call(
                        aws_service_name, resource_name
                    ),
                    "generic_resource_describe_call": hub.pop_create.aws.resource.build_get_by_resource_call(),
                }

                # get resource exec, state, and tool modules
                resource_modules["exec"] = {
                    f"{aws_service_name}.{resource_name}": hub.pop_create.aws.plugin.parse_exec_plugin(
                        ctx, shared_resource_data
                    )
                }

                resource_modules["state"] = {
                    f"{aws_service_name}.{resource_name}": hub.pop_create.aws.plugin.parse_state_plugin(
                        ctx, shared_resource_data
                    )
                }

                resource_modules["tool"] = {
                    f"{aws_service_name}.{resource_name}_utils": hub.pop_create.aws.utils.plugin.parse(
                        shared_resource_data
                    )
                }

                hub.pop_create.aws.init.run_pop_create_resource_modules(
                    ctx, directory, aws_service_name, resource_modules
                )

        # Now create any service related modules
        hub.pop_create.aws.init.run_pop_create_service_modules(
            ctx, directory, aws_service_name, session
        )

    return ctx


def plugin_exists(hub, ctx, aws_service_name: str, resource_name: str) -> bool:
    """
    Validate if the plugin path exists based on create plugin
    """
    path = pathlib.Path(ctx.target_directory).absolute() / ctx.clean_name
    if "auto_state" in ctx.create_plugin:
        path = path / "exec"
    elif "state_modules" in ctx.create_plugin:
        path = path / "states"
    elif "tests" in ctx.create_plugin:
        path = path / "tests" / "integration" / "states"

    path = path / aws_service_name / f"{resource_name}.py"
    if path.exists():
        hub.log.info(f"Plugin already exists at '{path}', use `--overwrite` to modify")
        return True

    return False


def run_pop_create_resource_modules(
    hub, ctx, directory: str, aws_service_name: str, modules: dict
):
    if ctx.create_plugin == "auto_state":
        # create exec and tools
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "auto_state", aws_service_name, modules["exec"]
        )
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "tool", aws_service_name, modules["tool"]
        )
        # TODO: run tests plugin
    elif ctx.create_plugin == "exec_modules":
        # create exec and tools
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "exec_modules", aws_service_name, modules["exec"]
        )
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "tool", aws_service_name, modules["tool"]
        )
        # TODO: run tests plugin
    elif ctx.create_plugin == "state_modules":
        # create exec, state and tools
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "exec_modules", aws_service_name, modules["exec"]
        )
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "state_modules", aws_service_name, modules["state"]
        )
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "tool", aws_service_name, modules["tool"]
        )
        # TODO: tests
    elif ctx.create_plugin == "tool":
        # create tool
        hub.pop_create.aws.init.create_plugin(
            ctx, directory, "tool", aws_service_name, modules["tool"]
        )


def run_pop_create_service_modules(
    hub,
    ctx,
    directory,
    aws_service_name: str,
    session: "boto3.session.Session",
):
    # Get service modules e.g. tag
    service_tag_methods = hub.pop_create.aws.service.parse_service_tag_methods(
        session=session,
        aws_service_name=aws_service_name,
    )

    tag_module = {
        f"{aws_service_name}.tag": hub.pop_create.aws.tag.plugin.parse(
            aws_service_name, service_tag_methods
        )
    }

    hub.pop_create.aws.init.create_plugin(
        ctx, directory, "tool", aws_service_name, tag_module
    )


def create_plugin(
    hub, ctx, directory, create_plugin_name, aws_service_name, plugins: dict
):
    try:
        # Initialize cloud spec and run it with provided create_plugin
        ctx.cloud_spec = NamespaceDict(
            api_version="",
            project_name=ctx.project_name,
            service_name=ctx.service_name,
            request_format=hub.pop_create.aws.init.resolve_plugin_request_format(
                create_plugin_name
            ),
            plugins=plugins,
        )

        hub.cloudspec.init.run(
            ctx,
            directory,
            create_plugins=[create_plugin_name],
        )
    finally:
        hub.log.info(
            f"Finished creating modules for service [{aws_service_name}] with create plugin {create_plugin_name}]"
        )
        # Reset it for the next resource or plugin
        ctx.cloud_spec = None


def resolve_plugin_request_format(hub, create_plugin_name):
    request_format = {}
    if create_plugin_name == "auto_state" or create_plugin_name == "exec_modules":
        request_format = {
            "get": hub.pop_create.aws.template.GET_REQUEST_FORMAT,
            "create": hub.pop_create.aws.template.CREATE_REQUEST_FORMAT,
            "delete": hub.pop_create.aws.template.DELETE_REQUEST_FORMAT,
            "update": hub.pop_create.aws.template.UPDATE_REQUEST_FORMAT,
            "list": hub.pop_create.aws.template.LIST_REQUEST_FORMAT,
        }
    elif create_plugin_name == "state_modules":
        request_format = {
            "present": hub.pop_create.aws.template.PRESENT_REQUEST_FORMAT,
            "absent": hub.pop_create.aws.template.ABSENT_REQUEST_FORMAT,
            "describe": hub.pop_create.aws.template.DESCRIBE_REQUEST_FORMAT,
        }
    elif create_plugin_name == "tool":
        request_format = {
            "get_tags_for_resource": hub.pop_create.aws.tag.template.GET_TAGS_FOR_RESOURCE_REQUEST,
            "update_tags": hub.pop_create.aws.tag.template.UPDATE_TAGS_REQUEST,
            "convert_raw_resource_to_present_async": hub.pop_create.aws.utils.template.CONVERT_RAW_RESOURCE_TO_PRESENT_REQUEST,
        }

    return request_format
